#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "charm++.h"
#include "util.h"
#include "jacobi.h"

int M; // readonly variable. All global variables must be readonly. 

#include "jacobi.def.h"

main::main(CkArgMsg *m)
{ 
  M = 5;
  A = CProxy_jacobi::ckNew(M); // 1-D decomposition: array of 5 cells
  CkStartQD(EntryIndex(main,allDone,CkQdMsg), &thishandle);
  CkPrintf("main done\n");
};

void main::allDone(CkQdMsg *m)
{
  CkPrintf("Quiescence detected. stopping\n");
  CkExit();
};

jacobi::jacobi(ArrayElementCreateMessage *msg) : ArrayElement(msg) 
{ 
 CkPrintf("[%d]: Element %d created\n", CkMyPe(), thisIndex); 
 int i = 10;
 int j= 100;
 data = initializeCell(thisIndex, 10, 100); // each cell 10x100

 count = 2;
 iterations = 0;
 sendToNeighbors();

 finishConstruction(); 
};

void jacobi::borderValues(DataMsg *m)
{
 CkPrintf("[%d] A[%d]: received message from nbr\n", CkMyPe(), thisIndex);

 copyBorder(data, (double *) &(m->b), m->fromLeft);

 if ((0 == --count)  && (++iterations < 3))
   {
     nextIteration(data);

     count = 2;
     sendToNeighbors();
   }
};

void
jacobi::sendToNeighbors(void)
{
 CkPrintf("[%d] A[%d]: sending to %d and %d\n", CkMyPe(), thisIndex,
	   (thisIndex+1)%M, (M+thisIndex-1)%M );
 CProxy_jacobi B(thisAID);

 DataMsg * d = new DataMsg;
 d->fromLeft = 0;
 fillBorder(data,  (double *) &(d->b), d->fromLeft);
 B[(thisIndex-1+M)%M].borderValues(d);

 d = new DataMsg;
 d->fromLeft = 1;
 fillBorder(data, (double *)&(d->b), d->fromLeft);
 B[(thisIndex+1)%M].borderValues(d);

};

jacobi::jacobi(ArrayElementMigrateMessage *m):
  ArrayElement(m)
{
  finishMigration();

};

// for migration, need to write pack interface.
